@@ -1,20 +1,13 @@
 #
 # DateEntry lets the user select a date from a drop-down calendar.
 #
-# Author: Hans J. Helgesen, October 1999.
-#
-# Maintainer is now Slaven Rezic <slaven@rezic.de>
-#
-# To contact the original author:
-# <hans.helgesen@novit.no>
-#
 # See end of this file for documentation.
 #
 package Tk::DateEntry;
 
-use vars qw($VERSION);
+use vars qw($VERSION $DEBUG);
 
-$VERSION = '1.38';
+$VERSION = '1.39';
 
 use Tk;
 use strict;
@@ -22,6 +15,13 @@ use Carp;
 use Time::Local;
 BEGIN { eval 'use POSIX qw(strftime)'; warn $@ if $@ }
 
+my $can_use_dcalc = 0;
+if (     eval q{ use Date::Calc qw(Day_of_Week check_date); 1 }) {
+    $can_use_dcalc = 1;
+} elsif (eval q{ use Date::Pcalc qw(Day_of_Week check_date); 1 }) {
+    $can_use_dcalc = 1;
+}
+
 require Tk::Frame;
 
 use base qw(Tk::Frame);
@@ -228,9 +228,9 @@ sub configure
     }
 
     if (defined($args{-dateformat}) &&
-	($args{-dateformat} < 1 || $args{-dateformat} > 3))
+	($args{-dateformat} < 1 || $args{-dateformat} > 4))
     {
-	carp ("-dateformat must be between 1 and 3");
+	carp ("-dateformat must be between 1 and 4");
 	delete $args{-dateformat};  # Ignore -dateformat
     }
 
@@ -240,12 +240,21 @@ sub configure
     if (defined($args{-daynames}) || defined($args{-weekstart})) {
 	# Refresh the daynames heading whenever -daynames or -weekstart
 	# changes.
-	my $daynames = $w->cget('-daynames');
+	my $daynames;
+	if ($args{-daynames} && $args{-daynames} eq 'locale' && defined &strftime) {
+	    foreach (0..6) {
+		push @$daynames, $w->_decode_posix_bytes(strftime("%a",0,0,0,1,1,1,$_));
+	    }
+	} else {
+	    $daynames = $w->cget('-daynames');
+	}
 	my $weekstart = $w->cget('-weekstart');
 
 	for (0..6) {
+	    my $dayname = $daynames->[($_ + $weekstart)%7];
+	    $dayname = $w->_bidify($dayname);
 	    $w->{_daylabel}->[$_]->configure
-		(-text => $daynames->[($_ + $weekstart)%7]);
+		(-text => $dayname);
 	}
     }
 }
@@ -288,11 +297,15 @@ sub buttonDown
 	#
 	my $cal = $w->getCalendar;
 
-	my $monthlabel = (defined &strftime
-			  ? strftime($w->cget('-headingfmt'),0,0,0,1,
-				     $w->{_month}-1,$w->{_year}-1900)
-			  : $w->{_month} . "/" . $w->{_year}
-			 );
+	my $monthlabel;
+	if (defined &strftime) {
+	    $monthlabel = strftime($w->cget('-headingfmt'),0,0,0,1,
+				   $w->{_month}-1,$w->{_year}-1900);
+	    $monthlabel = $w->_decode_posix_bytes($monthlabel);
+	} else {
+	    $monthlabel = $w->{_month} . "/" . $w->{_year};
+	}
+	$monthlabel = $w->_bidify($monthlabel);
 	$w->{_monthlabel}->configure(-text=>$monthlabel);
 
 	for my $week (0..5) {
@@ -481,6 +494,7 @@ sub defaultParse
     /^1$/ && (($m,$d,$y) = (split '/', $str));
     /^2$/ && (($y,$m,$d) = (split '/', $str));
     /^3$/ && (($d,$m,$y) = (split '/', $str));
+    /^4$/ && (($y,$m,$d) = (split '-', $str));
 
     return ($y,$m,$d);
 }
@@ -500,6 +514,8 @@ sub defaultFormat
 	sprintf("%04d/%02d/%02d", $y, $m, $d);
     } elsif (/^3$/) {
 	sprintf("%02d/%02d/%04d", $d, $m, $y);
+    } elsif (/^4$/) {
+	sprintf("%04d-%02d-%02d", $y, $m, $d);
     }
 }
 
@@ -521,6 +537,10 @@ sub getCalendar
 {
     my ($w) = @_;
 
+    if ($can_use_dcalc) {
+	return $w->getCalendarDC();
+    }
+
     my $week=0;
     my $cal=[];
 
@@ -547,6 +567,28 @@ sub getCalendar
     return $cal;
 }
 
+#--------------------
+# The same as getCalendar with Date::Calc
+#
+sub getCalendarDC
+{
+    my ($w) = @_;
+
+    my $week=0;
+    my $cal=[];
+    for my $mday (1..31) {
+	if (check_date($w->{_year},$w->{_month},$mday)) {
+	    my $wday = Day_of_Week($w->{_year},$w->{_month},$mday);
+	    $wday = ($wday - $w->cget('-weekstart')) % 7;
+	    $cal->[$week]->[$wday]=$mday;
+	    if ($wday == 6) {
+		$week++;
+	    }
+	}
+    }
+    return $cal;
+}
+
 #--------------------
 # Hide the window....
 #
@@ -627,57 +669,154 @@ sub prevMonth
 # increment or decrement the entry's day
 sub rotateDay
 {
-  my ( $w, $dir, $step ) = @_;
+    my ( $w, $dir, $step ) = @_;
 
-  $w->readContent;
-  $w->{_day} += $dir * $step;
-  # depend upon timelocal() to fix up days outside of 1..31
-  $w->updateDate;
+    $w->readContent;
+    $w->{_day} += $dir * $step;
+    # depend upon timelocal() to fix up days outside of 1..31
+    $w->updateDate;
 }
 
 sub rotateMonth
 {
-  my ( $w, $dir ) = @_;
+    my ( $w, $dir ) = @_;
 
-  $w->readContent;
+    $w->readContent;
 
-  $dir > 0 ? $w->nextMonth : $w->prevMonth;
+    $dir > 0 ? $w->nextMonth : $w->prevMonth;
 
-  $w->updateDate;
+    $w->updateDate;
 }
 
 sub rotateYear
 {
-  my ( $w, $dir ) = @_;
+    my ( $w, $dir ) = @_;
 
-  $w->readContent;
+    $w->readContent;
 
-  $w->{_year} += $dir;
+    $w->{_year} += $dir;
 
-  $w->updateDate;
+    $w->updateDate;
 }
 
 # make the entry widget track the internal notion of the date
 sub updateDate
 {
-  my ($w) = @_;
+    my ($w) = @_;
 
-  my ($d,$m,$y) = eval {
-    local $SIG{__DIE__} = undef;
-    (localtime(timelocal(0,0,0,
-			 $w->{_day},
-			 $w->{_month}-1,
-			 $w->{_year})))[3,4,5];
-  };
-  unless ( $@ )
-  {
-    $m++;
-    $y += 1900;
-    my $e = $w->Subwidget("entry");
-    $e->delete('0','end');
-    $e->insert('end',
-	       $w->Callback(-formatcmd=>$y,$m, $d));
-  }
+    if ($can_use_dcalc) {
+	return $w->updateDateDC();
+    }
+
+    my ($d,$m,$y) = eval {
+	local $SIG{__DIE__} = undef;
+	(localtime(timelocal(0,0,0,
+			     $w->{_day},
+			     $w->{_month}-1,
+			     $w->{_year})))[3,4,5];
+    };
+    unless ( $@ )
+	{
+	    $m++;
+	    $y += 1900;
+	    my $e = $w->Subwidget("entry");
+	    $e->delete('0','end');
+	    $e->insert('end',
+		       $w->Callback(-formatcmd=>$y,$m, $d));
+	}
+}
+
+sub updateDateDC
+{
+    my ($w) = shift;
+
+    my ($d,$m,$y);
+    if (check_date($w->{_year},$w->{_month},$w->{_day})) {
+	$y = $w->{_year};
+	$m = $w->{_month};
+	$d = $w->{_day};
+	my $e = $w->Subwidget("entry");
+	$e->delete('0','end');
+	$e->insert('end',
+		   $w->Callback(-formatcmd=>$y,$m, $d));
+    }
+}
+
+# "Stolen" from Locale::Maketext::Lexicon (called "encoding" there)
+# and slightly adapted.
+sub _posix_encoding {
+    my $w = shift;
+    if (exists $w->{_posix_encoding}) {
+	return $w->{_posix_encoding};
+    }
+
+    local $^W;    # no warnings 'uninitialized', really.
+    my ( $country_language, $locale_encoding );
+
+    local $@;
+    eval {
+        require I18N::Langinfo;
+        $locale_encoding =
+          I18N::Langinfo::langinfo( I18N::Langinfo::CODESET() );
+      }
+      or eval {
+        require Win32::OLE::NLS;
+	my $lcid = Win32::OLE::NLS::GetSystemDefaultLCID();
+        $locale_encoding = Win32::OLE::NLS::GetLocaleInfo($lcid, Win32::OLE::NLS::LOCALE_IDEFAULTANSICODEPAGE()); # XXX this is unchecked!
+	$locale_encoding = 'cp' . $locale_encoding if $locale_encoding;
+      };
+    if ( !$locale_encoding ) {
+        foreach my $key (qw( LANGUAGE LC_ALL LC_MESSAGES LANG )) {
+            $ENV{$key} =~ /^([^.]+)\.([^.:]+)/ or next;
+            ( $country_language, $locale_encoding ) = ( $1, $2 );
+            last;
+        }
+    }
+    if (   defined $locale_encoding
+        && lc($locale_encoding) eq 'euc'
+        && defined $country_language )
+    {
+        if ( $country_language =~ /^ja_JP|japan(?:ese)?$/i ) {
+            $locale_encoding = 'euc-jp';
+        }
+        elsif ( $country_language =~ /^ko_KR|korean?$/i ) {
+            $locale_encoding = 'euc-kr';
+        }
+        elsif ( $country_language =~ /^zh_CN|chin(?:a|ese)?$/i ) {
+            $locale_encoding = 'euc-cn';
+        }
+        elsif ( $country_language =~ /^zh_TW|taiwan(?:ese)?$/i ) {
+            $locale_encoding = 'euc-tw';
+        }
+    }
+
+    $w->{_posix_encoding} = $locale_encoding;
+}
+
+sub _decode_posix_bytes {
+    my($w, $string) = @_;
+    my $codeset = $w->_posix_encoding;
+    if ($codeset) {
+	eval {
+	    require Encode;
+	    $string = Encode::decode($codeset, $string);
+	};
+	warn "Cannot decode string '$string' in codeset '$codeset': $@" if $@;
+    }
+    $string;
+}
+
+sub _bidify {
+    my(undef, $string) = @_;
+    return if !defined $string;
+    eval {
+	if ($string =~ m{[\p{BidiClass:R}\p{BidiClass:AL}\p{BidiClass:RLE}\p{BidiClass:RLO}]}) {
+	    require Text::Bidi;
+	    $string = Text::Bidi::log2vis($string);
+	}
+    };
+    warn "Cannot align right-to-left text. $@" if $@ && $DEBUG;
+    $string;
 }
 
 1;
@@ -692,19 +831,19 @@ Tk::DateEntry - Drop down calendar widget for selecting dates.
 
 =head1 SYNOPSIS
 
-$dateentry = $parent->DateEntry (<options>);
+    $dateentry = $parent->DateEntry (<options>);
 
 =head1 DESCRIPTION
 
 Tk::DateEntry is a drop down widget for selecting dates. It looks like
 the BrowseEntry widget with an Entry followed by an arrow button, but
-in stead of displaying a Listbox the DateEntry displays a calendar
+instead of displaying a Listbox the DateEntry displays a calendar
 with buttons for each date. The calendar contains buttons for browsing
 through the months.
 
 When the drop down is opened, the widget will try to read the current
-content of the widget (the -textvariable), and display the month/year
-specified. If the variable is entry, or contains invalid data, the
+content of the widget (the B<-textvariable>), and display the month/year
+specified. If the variable is empty, or contains invalid data, then the
 current month is displayed. If one or two digit year is specified,
 the widget tries to guess the correct century by using a "100 year
 window".
@@ -736,18 +875,38 @@ currently selected.
 
 =head1 REQUIREMENTS
 
-Tk::DateEntry requires Time::Local and POSIX (strftime)
-(and basic Perl/Tk of course....)
+Tk::DateEntry requires L<Time::Local> and L<POSIX> (strftime) (and
+basic Perl/Tk of course....). For using dates before 1970-01-01 either
+L<Date::Calc> or L<Date::Pcalc> is required.
+
+For faster scanning between months the optional requirement
+L<Tk::FireButton> is needed. For localized day and month names the
+following modules are needed:
+
+=over
+
+=item L<Encode> (for non-ASCII encodings)
+
+=item L<I18N::Langinfo> (Unix-like systems only, for detection of
+current locale)
+
+=item L<Win32::OLE::NLS> (Windows only, for detection of current
+locale)
+
+=item L<Text::Bidi> (only if support for right-to-left scripts
+(Hebrew, Arabic) is needed)
+
+=back
 
 =head1 OPTIONS
 
 =over 4
 
-=item -arrowimage => image
+=item B<-arrowimage> => I<image>
 
 Use alternative image for the arrow button.
 
-=item -dateformat => integer
+=item B<-dateformat> => I<integer>
 
 Specify dateformat to use:
 
@@ -765,95 +924,102 @@ Specify dateformat to use:
 
 3 = DD/MM/YYYY
 
+=item
+
+4 = YYYY-MM-DD
+
 =back
 
-See also "DATE FORMATS" below.
+See also L</DATE FORMATS> below.
 
-=item -parsecmd => \&callback
+=item B<-parsecmd> => I<\&callback>
 
 Instead of using one of the builtin dateformats, you can specify your
-own by supplying a subroutine for parsing (-parsecmd) and formatting
-(-formatcmd) of the date string. These options overrides -dateformat.
-See "DATE FORMATS" below.
+own by supplying a subroutine for parsing (B<-parsecmd>) and formatting
+(B<-formatcmd>) of the date string. These options overrides B<-dateformat>.
+See L</DATE FORMATS> below.
 
-=item -formatcmd => \&callback
+=item B<-formatcmd> => I<\&callback>
 
-See -parsecmd above and "DATE FORMATS" below.
+See B<-parsecmd> above and L</DATE FORMATS> below.
 
-=item -background => color
+=item B<-background> => I<color>
 
 Sets the background color for the Entry subwidget. Note that
 the dropdown calendar is not affected by this option. See also
--boxbackground, -buttonbackground and -todaybackground.
+B<-boxbackground>, B<-buttonbackground> and B<-todaybackground>.
 
-=item -buttonbackground => color
+=item B<-buttonbackground> => I<color>
 
 Sets the background color for all button in the dropdown calendar.
 
-=item -boxbackground => color
+=item B<-boxbackground> => I<color>
 
 Sets the background color for the dropdown widget (not including the buttons).
 
-=item -todaybackground => color
+=item B<-todaybackground> => I<color>
 
 Sets the background color for the button representing the current date.
 
-=item -font => font
+=item B<-font> => I<font>
 
 Sets the font for all subwidgets.
 
-=item -configcmd => \&callback
+=item B<-configcmd> => I<\&callback>
 
 Called for every day button in the calendar while month configuration.
 A hash with the keys B<-date>, B<-widget>, and B<-datewidget> is
-passed to the callback. The -date parameter is an array reference
+passed to the callback. The B<-date> parameter is an array reference
 containing day, month, and year. For empty buttons this parameter is
-undefined. The -widget parameter is a reference to the current
-Tk::DateEntry widget, and the -datewidget parameter is a reference to
+undefined. The B<-widget> parameter is a reference to the current
+Tk::DateEntry widget, and the B<-datewidget> parameter is a reference to
 the current day button. A sample callback:
 
     sub configcmd {
        my(%args) = @_;
-       my($day,$month,$year) = @{ $args{-date} };
-       my $widget = $args->{-widget};
-       my $datewidget = $args->{-datewidget};
+       my($day,$month,$year) = @{$args->{-date}};
+       my $widget            =   $args->{-widget};
+       my $datewidget        =   $args->{-datewidget};
        $datewidget->configure(...);
        ...
     }
 
 The callback is called after initial configuration of a day widget,
 that is, i.e. the label and the background color is already set. Note
-that day buttons XXX beibehalten XXX their configuration while
-switching between months.
+that day buttons keep their configuration while switching between
+months.
 
-=item -daynames => [qw/Sun Mon Tue Wed Thu Fri Sat/]
+=item B<-daynames> => [qw/Sun Mon Tue Wed Thu Fri Sat/]
 
 Specifies the daynames which is used in the calendar heading.
-The default is [qw/S M Tu W Th F S/]. Note that the array MUST begin
-with the name of Sunday, even if -weekstart specifies something else
-than 0 (which is Sunday). See also "WEEKS" below.
+The default is C<< [qw/S M Tu W Th F S/] >>. Note that the array MUST begin
+with the name of Sunday, even if B<-weekstart> specifies something else
+than 0 (which is Sunday). See also L</WEEKS> below.
 
-=item -weekstart => number
+It is also possible to use the special value C<locale> to use the
+daynames from the current locale.
+
+=item B<-weekstart> => I<number>
 
 Use this if you don't want the weeks to start on Sundays. Specify a number
-between 0 (Sunday) and 6 (Saturday). See "WEEKS" below.
+between 0 (Sunday) and 6 (Saturday). See L</WEEKS> below.
 
-=item -headingfmt => string
+=item B<-headingfmt> => I<string>
 
-Format for the Month name heading. The month name heading is created by
-calling strftime(format,0,0,0,1,month,year). Default format is '%B %Y'.
+Format for the month name heading. The month name heading is created by
+calling C<< strftime(format,0,0,0,1,month,year) >>. Default format is 'C<< %B %Y >>'.
 Note that only month and year will have sensible values, including
 day and/or time in the heading is possible, but it makes no sense.
 
-If POSIX.pm is not available then this option has no effect and the
-month name heading format will just be "%m/%Y".
+If L<POSIX> is not available then this option has no effect
+and the month name heading format will just be "C<< %m/%Y >>".
 
-=item -state => string
+=item B<-state> => I<string>
 
-'normal', 'disabled' or 'readonly'. The latter forces the user to use
+'B<normal>', 'B<disabled>' or 'B<readonly>'. The latter forces the user to use
 the drop down, editing in the Entry subwidget is disabled.
 
-=item -width => number
+=item B<-width> => I<number>
 
 Width of the Entry subwidget, default is 10 (which fits the default
 date format MM/DD/YYYY).
@@ -884,6 +1050,10 @@ The "builtin" date formats are:
 
 -dateformat => 3       - DD/MM/YYYY
 
+=item
+
+-dateformat => 4       - YYYY-MM-DD
+
 =back
 
 Trailing fields that are missing will be replaced by the current date, if the
@@ -891,34 +1061,34 @@ year is specified by one or two digits, the widget will guess the century by
 using a "100 year window".
 
 If you're not satisified with any of these formats, you might specify your
-own parse- and format routine by using the -parsecmd and -formatcmd options.
+own parse and format routine by using the B<-parsecmd> and B<-formatcmd> options.
 
-The -parsecmd subroutine will be called whenever the pulldown is opened.
-The subroutine will be called with the current content of -textvariable as
+The B<-parsecmd> subroutine will be called whenever the pulldown is opened.
+The subroutine will be called with the current content of B<-textvariable> as
 the only argument. It should return a three element list: (year, month, day).
 Any undefined elements will be replaced by default values.
 
-The -formatcmd subroutine will be called whenever the user selects a date.
+The B<-formatcmd> subroutine will be called whenever the user selects a date.
 It will be called with three arguments: (year, month, day). It should return
-a single string which will be assigned to the -textvariable.
+a single string which will be assigned to the B<-textvariable>.
 
-See "EXAMPLES" below.
+See L</EXAMPLES> below.
 
 =head1 WEEKS
 
-The default is to display the calendar the same way as the unix "cal" command
+The default is to display the calendar the same way as the unix L<cal(1)> command
 does: Weeks begin on Sunday, and the daynames are S, M, Tu, W, Th, F, and S.
 
 However, some people prefer to start the weeks at Monday (saving both
 Saturday and Sunday to the weekEND...)  This can be achived by specifying
--weekstart=>1. -weekstart=>0 causes the week to start at Sunday, which
+C<< -weekstart=>1 >>. C<< -weekstart=>0 >> causes the week to start at Sunday, which
 is the default. If you have a very odd schedule, you could also start the
-week at Wednesday by specifying -weekstart=>3 .....
+week at Wednesday by specifying C<< -weekstart=>3 >>.....
 
 If you don't like the "cal" headings, you might specify something else
-by using the -daynames option.
+by using the B<-daynames> option.
 
-See "EXAMPLES" below.
+See L</EXAMPLES> below.
 
 =head1 EXAMPLES
 
@@ -930,6 +1100,10 @@ See "EXAMPLES" below.
 
 If you want the "locale's abbreviated weekday name" you do it like this:
 
+	$parent->DateEntry(-daynames=>'locale')->pack;
+
+which is short for:
+
 	use POSIX qw/strftime/;
 	my @daynames=();
 	foreach (0..6) {
@@ -953,20 +1127,34 @@ A Norwegian would probably do something like this:
 		 }
 		)->pack;
 
-Note that this -parsecmd will return (undef,undef,undef) even if one or
+Note that this B<-parsecmd> will return (undef,undef,undef) even if one or
 two of the fields are present. A more sophisticated regex might be needed....
 
 =head1 CAVEATS
 
-Tk::DateEntry uses timelocal(), localtime() and strftime().
-These functions are based on the standard unix time representation, which
-is the number of seconds since 1/1/1970.
+If neither L<Date::Calc> nor L<Date::Pcalc> are available, then
+Tk::DateEntry uses timelocal(), localtime() and strftime(). These
+functions are based on the standard unix time representation, which is
+the number of seconds since 1970-01-01. This means that in this case
+Tk::DateEntry doesn't support dates prior to 1970, and on a 32 bit
+computer it doesn't support dates after 2037-12-31.
 
-This means that Tk::DateEntry doesn't support dates prior to 1970, and on
-a 32 bit computer it doesn't support dates after 12/31/2037.
+Future perl versions (possibly beginning with 5.10.1) will have
+support for 64 bit times.
+
+=head1 TODO
+
+Use L<DateTime::Locale> instead of L<POSIX> for localized day and
+month names.
 
 =head1 SEE ALSO
 
-L<Tk::Entry>, L<Tk::Button>.
+L<Tk::Entry>, L<Tk::Button>, L<Tk::ChooseDate>.
+
+=head1 AUTHOR
+
+Hans J. Helgesen <hans.helgesen@novit.no>, October 1999.
+
+Current maintainer is Slaven Rezic <slaven@rezic.de>.
 
 =cut
